#include <stdio.h>
#include <pthread.h>
#include <math.h>
#include "prover.h"
#include "symbol_table.h"
#include "clauses.h"


#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))

extern pthread_mutex_t mutexprops;
extern int diamond_restricted_res;

extern int print_queries;
extern int print_subsumed;
extern int numagents;

extern void free_clauses_set_p(hml_pclauses **set);

extern clauseslist *build_modal_candidates(hml_mclauses *set, int agent, int
modal_level, int type, int literal);

// extern literalslist* check_sat_clause(int component, literalslist *clause);
extern int make_query(int component, literalslist *literals);
extern void bprint(char **buffer, char *string);


extern int forward_subsumed (tclause **clause, hml_pclauses *set);

extern void s_print_mclauses (char **s, hml_mclauses *set);
extern void s_print_pclauses (char **s, hml_pclauses *set);
extern void print_pclauses (hml_pclauses *set);
extern void print_query(tclause *clause, int component, int sat);

extern tclause* to_tclause(literalslist *literals, int rule);
extern void append_clause_set(tclause **clause, hml_pclauses **set);

extern void print_deleted_clauses (void);
extern void print_clauses (void);
extern void print_clause(tclause *clause);
extern void print_disjunction(literalslist *s);
extern void s_print_disjunction(char** string, literalslist *s);
extern literalslist *free_disjunction(literalslist *dis);
extern void delete_clause_prop (hml_pclauses **set,tclause *c);
extern void delete_clause_modal (hml_mclauses **set, tclause *c);
extern clauseslist *delete_clause_list(tclause *c, clauseslist *l);
extern void insert_clause_deleted (tclause *clause);

extern agent_node *find_agent (int id);

extern int clause_in_list(tclause *c, clauseslist *l);

extern void insert_clause_set (tclause **clause, int type, int set);
extern int insert_clause_prop (tclause **clause, hml_pclauses **set, int set_flag);
extern void delete_clause (tclause *clause);
extern void delete_clause_all(int number);
extern clauseslist *remove_clause_list (tclause *c, clauseslist *l);
extern int size_clause (literalslist *l);
extern int max_lit_clause (literalslist *l);
extern int classify_clause (literalslist *l);

extern tclause *get_clause (int clause);
extern literalslist *simplify_clause (literalslist *l);
extern void print_out (char *);

extern int modal_level_elimination (void);
extern int pure_literal_elimination (void);
extern int modal_level_ple (void);
extern int unit_resolution (hml_pclauses **);

void forward_subsumes (tclause *clause, hml_pclauses **set);
int forward_subsumed (tclause **clause, hml_pclauses *set);
void backward_subsumes (tclause *clause, hml_pclauses **set);
int backward_subsumed (tclause **clause, hml_pclauses *set);

extern void do_lres (int clause);
extern void do_gen3 (int clause, hml_mclauses *m_set);
extern void do_gen1 (int clause, hml_mclauses *m_set);
extern void do_ires (int clause);
extern void do_mod_ordires (int clause);

extern pthread_mutex_t mutexallclauses;

extern hml_pclauses *l_sos;
extern hml_pclauses *l_usable;
extern hml_pclauses *l_sos_1;
extern hml_pclauses *l_usable_1;
extern hml_pclauses *l_sos_2;
extern hml_pclauses *l_usable_2;
extern hml_pclauses *i_sos;
extern hml_pclauses *i_usable;

extern hml_mclauses *m_sos;
extern hml_mclauses *m_usable;
extern hml_mclauses *m_sos_1;
extern hml_mclauses *m_usable_1;
extern hml_mclauses *m_sos_2;
extern hml_mclauses *m_usable_2;

extern int print_generated;
extern int cycles;
extern int ires;
extern int global;
extern int modal_positive;
extern int modal_negative;
extern int strong_modal_positive;
extern int strong_modal_negative;
extern int literal_selection;
extern int numclause;
extern int satmle;
extern int mle;
extern int ple;
extern int ml_ple;
extern int unit;
extern int lhs_unit;
extern int forward_subsumption;
extern int backward_subsumption;
extern int sos_subsumption;
extern int order_level;
extern int clause_selection;
extern int maxproof;
extern int numproofs;
extern int numsat;
extern int numpropagateunique;
extern int mdepth;
extern int propagate_box;
extern int propagate_dia;
extern int interleave_ires;

literalslist* get_literals (hml_mclauses *set) {
    hml_mclauses *aux = set;
    hagentclauses *hag = NULL;
    htypeclauses *htc = NULL;
    hmaxlitclauses *hmaxlit = NULL;
    literalslist *head = NULL;

    for (aux=set; aux != NULL; aux=aux->hml.next)
        for (hag=aux->id; hag != NULL; hag=hag->hid.next)
            for (htc=hag->type; htc != NULL; htc=htc->htype.next) {
                agent_node *a = find_agent(hag->id);
                if(a != NULL && strstr(a->name, "dbox") != NULL) {
                    for (hmaxlit = htc->lit; hmaxlit != NULL; hmaxlit=hmaxlit->hmax.next) {
                        clauseslist *l = hmaxlit->list;
                        while (l != NULL) {
                            tclause *clause = get_clause(l->clause_number);
                            literalslist *n = malloc(sizeof(literalslist));
                            n->literal = clause->right->literal;
                            n->next = head;
                            head = n;
                            l = l->next;
                        }
                    }
                }
            }

    return head;
}

void move_clause_modal(tclause *c, hml_mclauses **set_out, hml_mclauses **set_in) {

    hml_mclauses *hmlm = NULL;
    hagentclauses *hag = NULL;
    htypeclauses *htc = NULL;
    hmaxlitclauses *hmaxlit = NULL;

    pthread_mutex_lock(&mutexprops);
    HASH_FIND(hml,*set_out,&(c->modal_level),sizeof(int),hmlm);
    if (hmlm != NULL) {
        HASH_FIND(hid,hmlm->id,&(c->agent),sizeof(int),hag);
        if (hag != NULL) {
            HASH_FIND(htype,hag->type,&(c->type),sizeof(int),htc);
            if (htc != NULL) {
                HASH_FIND(hmax,htc->lit,&(c->max_literal),sizeof(int),hmaxlit);
                if (hmaxlit != NULL) {
                    hmaxlit->list = remove_clause_list(c,hmaxlit->list);
                    if (hmaxlit->list == NULL) {
                        HASH_DELETE(hmax,htc->lit,hmaxlit);
                        free(hmaxlit);
                        hmaxlit = NULL;
                    }
                }
                if (htc->lit == NULL) {
                    HASH_DELETE(htype,hag->type,htc);
                    free(htc);
                    htc = NULL;
                }
            }

            if (hag->type == NULL) {
                HASH_DELETE(hid,hmlm->id,hag);
                free(hag);
                hag = NULL;
            }
        }
        if (hmlm->id == NULL) {
            HASH_DELETE(hml,*set_out,hmlm);
            free(hmlm);
            hmlm = NULL;
        }
    }

    HASH_FIND(hml,*set_in,&(c->modal_level),sizeof(int),hmlm);
    if (hmlm == NULL) {
        hml_mclauses *new1 = malloc(sizeof(*new1));
        new1->ml = c->modal_level;
        new1->id = NULL;
        HASH_ADD(hml,*set_in,ml,sizeof(int),new1);
        hmlm = new1;
    }
    HASH_FIND(hid,hmlm->id,&(c->agent),sizeof(int),hag);
    if (hag == NULL) {
        hagentclauses *new2 = malloc(sizeof(*new2));
        new2->id = c->agent;
        new2->type = NULL;
        HASH_ADD(hid,hmlm->id,id,sizeof(int),new2);
        hag = new2;
    }
    HASH_FIND(htype,hag->type,&(c->type),sizeof(int),htc);
    if (htc == NULL) {
        htypeclauses *new3 = malloc(sizeof(*new3));
        new3->type = c->type;
        new3->lit = NULL;
        HASH_ADD(htype,hag->type,type,sizeof(int),new3);
        htc = new3;
    }
    HASH_FIND(hmax,htc->lit,&(c->max_literal),sizeof(int),hmaxlit);
    if (hmaxlit == NULL) {
        hmaxlitclauses *new4 = malloc(sizeof(*new4));
        new4->max = c->max_literal;
        new4->list = NULL;
        HASH_ADD(hmax,htc->lit,max,sizeof(int),new4);
        hmaxlit = new4;
    }    
    clauseslist *new = malloc(sizeof(clauseslist));
    new->clause_number = c->number;
    new->next = hmaxlit->list;
    hmaxlit->list = new;
    pthread_mutex_unlock(&mutexprops);
}

void move_clause_propositional(tclause *c, hml_pclauses **set_out, hml_pclauses **set_in) {

    hml_pclauses *hmlp = NULL;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;

    pthread_mutex_lock(&mutexprops);
    HASH_FIND(hml,*set_out,&(c->modal_level),sizeof(int),hmlp);
    if (hmlp != NULL) {
        HASH_FIND(hmax,hmlp->max,&(c->max_literal),sizeof(int),hmc);
        if (hmc != NULL) {
            HASH_FIND(hsize,hmc->size,&(c->size),sizeof(int),hsc);
            if (hsc != NULL) {
                hsc->list = remove_clause_list(c,hsc->list);
                if (hsc->list == NULL) {
                    HASH_DELETE(hsize,hmc->size,hsc);
                    free(hsc);
                    hsc = NULL;
                }
            }
            if (hmc->size == NULL) {
                HASH_DELETE(hmax,hmlp->max,hmc);
                free(hmc);
                hmc = NULL;
            }
        }
        if (hmlp->max == NULL) {
            HASH_DELETE(hml,*set_out,hmlp);
            free(hmlp);
            hmlp = NULL;
        }
    }

    HASH_FIND(hml,*set_in,&(c->modal_level),sizeof(int),hmlp);
    if (hmlp == NULL) {
        hml_pclauses *new1 = malloc(sizeof(*new1));
        new1->ml = c->modal_level;
        new1->list_selection = NULL;
        new1->last_selection = NULL;
        new1->max = NULL;
        HASH_ADD(hml,*set_in,ml,sizeof(int),new1);
        hmlp = new1;
    }
    HASH_FIND(hmax,hmlp->max,&(c->max_literal),sizeof(int),hmc);
    if (hmc == NULL) {
        hmaxclauses *new2 = malloc(sizeof(*new2));
        new2->max = c->max_literal;
        new2->size = NULL;
        HASH_ADD(hmax,hmlp->max,max,sizeof(int),new2);
        hmc = new2;
    }
    HASH_FIND(hsize,hmc->size,&(c->size),sizeof(int),hsc);
    if (hsc == NULL) {
        hsizeclauses *new3 = malloc(sizeof(*new3));
        new3->size = c->size;
        new3->list = NULL;
        HASH_ADD(hsize,hmc->size,size,sizeof(int),new3);
        hsc = new3;
    }

    clauseslist *new = malloc(sizeof(clauseslist));
    new->clause_number = c->number;
    new->next = hsc->list;
    hsc->list = new;
    pthread_mutex_unlock(&mutexprops);
}

int check_in_sos(tclause *clause, hml_pclauses *l_sos) {
    hml_pclauses *hmlp = NULL;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;

    int in = 0;
    if (clause != NULL) {
        HASH_FIND(hml,l_sos,&(clause->modal_level),sizeof(int),hmlp);
        if (hmlp != NULL) {
            HASH_FIND(hmax,hmlp->max,&(clause->max_literal),sizeof(int),hmc);
            if (hmc != NULL) {
                HASH_FIND(hsize,hmc->size,&(clause->size),sizeof(int),hsc);
                if (hsc != NULL) {
                    clauseslist *aux = hsc->list;
                    while (aux != NULL && !in) {
                        if (aux->clause_number == clause->number)
                            in = 1;
                        else aux = aux->next;
                    }
                }
            }
        }
    }
    return in;
}


int chosen (hml_pclauses **hmlp, hml_pclauses *l_sos) {
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL; 

    pthread_mutex_lock(&mutexprops);
    if ((*hmlp) != NULL &&
            (*hmlp)->max != NULL &&
            (*hmlp)->max->size != NULL &&
            (*hmlp)->max->size->list != NULL) {
        int clause = (*hmlp)->max->size->list->clause_number;
        int size_clause = (*hmlp)->max->size->size;
        int max_lit_clause = (*hmlp)->max->max;
        if (clause_selection == SHORTEST) {
            for (hmc=(*hmlp)->max; hmc != NULL; hmc = hmc->hmax.next) {
                for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                    if (hsc->size < size_clause) {
                        size_clause = hsc->size;
                        clause = hsc->list->clause_number;
                    }
            }
            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
        else if (clause_selection == SMALLEST) {
            for (hmc=(*hmlp)->max; hmc != NULL; hmc = hmc->hmax.next) {
                if (hmc->max < max_lit_clause) {
                    max_lit_clause = hmc->max;
                }
            }
            HASH_FIND(hmax,(*hmlp)->max,&(max_lit_clause),sizeof(int),hmc);
            for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                if (hsc->size < size_clause) {
                    size_clause = hsc->size;
                    clause = hsc->list->clause_number;
                }

            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
        else if (clause_selection == GREATEST) {
            for (hmc=(*hmlp)->max; hmc != NULL; hmc = hmc->hmax.next) {
                if (hmc->max > max_lit_clause) {
                    max_lit_clause = hmc->max;
                }
            }
            HASH_FIND(hmax,(*hmlp)->max,&(max_lit_clause),sizeof(int),hmc);
            for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                if (hsc->size < size_clause) {
                    size_clause = hsc->size;
                    clause = hsc->list->clause_number;
                }

            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
        else if (clause_selection == NEWEST || clause_selection == OLDEST) {
            clauseslist *aux = (*hmlp)->list_selection;
            if (aux != NULL) {
                tclause *clause = get_clause(aux->clause_number);
                int in_sos = check_in_sos(clause, l_sos);
                while (aux!= NULL && (clause == NULL || !in_sos)) { // that means it is not in the allclauses hash, so it was deleted
                    clauseslist *aux2 = aux;
                    aux = aux->next;
                    free(aux2);
                    clause = get_clause(aux->clause_number);
                    in_sos = check_in_sos(clause, l_sos);
                }
                if (aux != NULL) {
                    (*hmlp)->list_selection = aux->next;
                    free(aux);
                }
                if ((*hmlp)->list_selection == NULL) {
                    if ((*hmlp)->last_selection != NULL) {
                        (*hmlp)->last_selection = NULL;
                    }
                }
                if (clause != NULL) {
                    pthread_mutex_unlock(&mutexprops);
                    return clause->number;
                }
            }
        }
    }
    pthread_mutex_unlock(&mutexprops);
    return 0;
}

int chosen_level (hml_pclauses **hmlp) {
    hmaxclauses *hmax = (*hmlp)->max;
    hmaxclauses *hmc = hmax;
    hsizeclauses *hsc = NULL; 

    pthread_mutex_lock(&mutexprops);
    if (hmc != NULL &&
            hmc->size != NULL &&
            hmc->size->list != NULL) {
        int clause = hmc->size->list->clause_number;
        int size_clause = hmc->size->size;
        int max_lit_clause = hmc->max;
        if (clause_selection == SHORTEST) {
            for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                if (hsc->size < size_clause) {
                    size_clause = hsc->size;
                    clause = hsc->list->clause_number;
                }
            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
        else if (clause_selection == SMALLEST) {
            for (hmc=hmax; hmc != NULL; hmc = hmc->hmax.next) {
                if (hmc->max < max_lit_clause) {
                    max_lit_clause = hmc->max;
                }
            }
            HASH_FIND(hmax,hmax,&(max_lit_clause),sizeof(int),hmc);
            for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                if (hsc->size < size_clause) {
                    size_clause = hsc->size;
                    clause = hsc->list->clause_number;
                }

            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
        else if (clause_selection == GREATEST) {
            for (hmc=hmax; hmc != NULL; hmc = hmc->hmax.next) {
                if (hmc->max > max_lit_clause) {
                    max_lit_clause = hmc->max;
                }
            }
            HASH_FIND(hmax,hmax,&(max_lit_clause),sizeof(int),hmc);
            for (hsc = hmc->size; hsc != NULL; hsc = hsc->hsize.next)
                if (hsc->size < size_clause) {
                    size_clause = hsc->size;
                    clause = hsc->list->clause_number;
                }

            pthread_mutex_unlock(&mutexprops);
            return clause;
        }
    }
    pthread_mutex_unlock(&mutexprops);
    return 0;  
}


void propagate_unique_dia_ml (hml_mclauses *m_sos) {

    hml_mclauses *aux = NULL;
    hagentclauses *hag = NULL;

    aux = m_sos;
    while (aux != NULL) {
        hag = aux->id;
        aux = aux->hml.next;
        while (hag != NULL) {
            htypeclauses *htcneg = NULL;
            if (hag->type->type == MPOSITIVE) {
                htcneg = hag->type->htype.next;
            }
            else {
                htcneg = hag->type;
            }

            hag = hag->hid.next;
            if (htcneg != NULL) {

                hmaxlitclauses *auxlit = htcneg->lit;

                int size = 0;
                while (auxlit != NULL) {
                    clauseslist *auxclauses = auxlit->list;
                    while (auxclauses != NULL) {
                        size++;
                        auxclauses = auxclauses->next;
                    }
                    auxlit = auxlit->hmax.next;
                }

                if (size == 1) { // || diamond_restricted_res == ON) {

                    int clause = htcneg->lit->list->clause_number;
                    tclause *c = get_clause(clause);

                    if (c->right->literal != CTRUE && !(c->propagated)) {
                        numpropagateunique++;
                        c->propagated = 1;
                        tclause *c1 = malloc(sizeof(tclause));
                        c1->number = numclause++;
                        c1->type = LITERAL;
                        c1->subtype = c->numdiamond;
                        c1->numdiamond = BOX;
                        c1->agent = 0;
                        c1->modal_level = c->modal_level + 1;
                        c1->left = CTRUE;
                        c1->right = malloc(sizeof(literalslist));
                        c1->right->literal = c->right->literal;
                        c1->right->next = NULL;
                        justification *just1 = malloc(sizeof(justification));
                        just1->rule = PROPUNIQUEDIA;
                        just1->parents = malloc(sizeof(clauseslist));
                        just1->parents->clause_number = c->number;
                        just1->parents->next = NULL;
                        just1->literals = NULL;
                        c1->just = just1;
                        c1->deleted = NULL;
                        c1->size = size_clause(c1->right);
                        c1->max_literal = max_lit_clause(c1->right);
                        c1->min_literal = c1->right->literal;
                        c1->class = classify_clause(c1->right);
                        insert_clause_set(&c1,c1->type,SOS);
                    }
                }
            }
        }
    }
}


void propagate_unique_box_ml (hml_mclauses *m_sos) {

    hml_mclauses *aux = NULL, *auxtmp = NULL;
    hagentclauses *hag = NULL, * hagtmp = NULL;

    HASH_ITER(hml,m_sos,aux,auxtmp) {
        HASH_ITER(hid,aux->id,hag,hagtmp) {
            htypeclauses *htcpos = NULL;

            if (hag->type->type == MPOSITIVE) {
                htcpos = hag->type;
            }
            else {
                htcpos = hag->type->htype.next;
            }

            if (htcpos != NULL) {
                hmaxlitclauses *auxlit = htcpos->lit;
                int size = 0;
                while (auxlit != NULL) {
                    clauseslist *auxclauses = auxlit->list;
                    while (auxclauses != NULL) {
                        size++;
                        auxclauses = auxclauses->next;
                    }
                    auxlit = auxlit->hmax.next;
                }

                if (size == 1) {

                    int clause = htcpos->lit->list->clause_number;
                    tclause *c = get_clause(clause);

                    if (c->right->literal != CFALSE) {
                        numpropagateunique++;
                        tclause *c1 = malloc(sizeof(tclause));
                        c1->number = numclause++;
                        c1->type = LITERAL;
                        c1->subtype = BOX;
                        c1->agent = 0;
                        c1->modal_level = c->modal_level + 1;
                        c1->left = CTRUE;
                        c1->right = malloc(sizeof(literalslist));
                        c1->right->literal = c->right->literal;
                        c1->right->next = NULL;
                        justification *just1 = malloc(sizeof(justification));
                        just1->rule = PROPUNIQUEBOX;
                        just1->parents = malloc(sizeof(clauseslist));
                        just1->parents->clause_number = c->number;
                        just1->parents->next = NULL;
                        just1->literals = NULL;
                        c1->just = just1;
                        c1->deleted = NULL;
                        c1->size = size_clause(c1->right);
                        c1->max_literal = max_lit_clause(c1->right);
                        c1->min_literal = c1->right->literal;
                        c1->class = classify_clause(c1->right);
                        insert_clause_set(&c1,c1->type,SOS);
                    }
                }
            }
        }
    }
}

int ml_srt_lt(hml_pclauses *a, hml_pclauses *b) {
    return (a->ml < b->ml);
}

int ml_srt_gt(hml_pclauses *a, hml_pclauses *b) {
    return (a->ml > b->ml);
}

hml_pclauses *shuffle(hml_pclauses *levels) {
    hml_pclauses *fast = NULL;
    hml_pclauses *slow = NULL;

    //  printf("\n begin shuffle \n");
    slow = levels;
    if (levels != NULL)
        fast = levels->hml.next;
    while (fast != NULL) {
        fast = fast->hml.next;
        if (fast != NULL) {
            slow = slow->hml.next;
            fast = fast->hml.next;
        }
    }
    if (slow != NULL)
        fast = slow->hml.next;
    slow->hml.next = NULL;
    slow = fast;
    fast = levels;

    /*printf("\n splits (slow):");
      hml_pclauses *aux,*tmp;
      HASH_ITER(hml,slow,aux,tmp) {
      printf("\n %d", aux->ml);
      }
      printf("\n splits (fast):");
      HASH_ITER(hml,fast,aux,tmp) {
      printf("\n %d", aux->ml);
      }


      printf("\n end split \n");
      */
    /* merge */
    hml_pclauses *new = NULL;
    while (slow != NULL) {
        hml_pclauses *slownext = slow->hml.next;
        HASH_ADD(hml,new,ml,sizeof(int),slow);
        //printf("\n adding slow %d",slow->ml);
        slow = slownext;
        if (fast != NULL) {
            hml_pclauses *fastnext = fast->hml.next;
            HASH_ADD(hml,new,ml,sizeof(int),fast);
            //printf("\n adding fast %d",fast->ml);
            fast = fastnext;
        }
    }
    //  printf("end while");
    while (fast != NULL) {
        HASH_ADD(hml,new,ml,sizeof(int),fast);
        fast = fast->hml.next;
    }

    /*HASH_ITER(hml,new,aux,tmp) {
      printf("\n %d", aux->ml);
      }
      printf("\n end shuffle \n"); */
    return new;
}




int move_pclauses_from_set (hml_pclauses **orig, int set_flag) {
    hml_pclauses *aux = *orig;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;  
    clauseslist *stored = NULL;
    int moved = 0;

    for (aux= *orig; aux != NULL; aux=aux->hml.next) {
        for (hmc=aux->max; hmc != NULL; hmc=hmc->hmax.next) {
            for (hsc=hmc->size; hsc != NULL; hsc=hsc->hsize.next) {
                clauseslist *l = hsc->list;
                while (l != NULL) {
                    tclause *clause = get_clause(l->clause_number);
                    clauseslist *auxcl = malloc(sizeof(clauseslist));
                    auxcl->clause_number = l->clause_number;
                    auxcl->clause_pointer = clause;
                    auxcl->next = stored;
                    stored = auxcl;

                    l = l->next;
                }
            }
        }
    }

    while(stored) {
        delete_clause_prop(orig, stored->clause_pointer);
        if (set_flag == SOS) {
            moved += insert_clause_prop(&(stored->clause_pointer), &l_sos, set_flag);
        } else if (set_flag == SOS_1) {
            moved += insert_clause_prop(&(stored->clause_pointer), &l_sos_1, set_flag);
        } else if (set_flag == SOS_2) {
            moved += insert_clause_prop(&(stored->clause_pointer), &l_sos_2, set_flag);
        } else {
            printf("\n ERROR move_pclauses: Invalid set_flag.\n");
            exit(1);
        } 
        clauseslist *d = stored;
        stored = stored->next;
        free(d);

    }


    return moved;
}


/**
 * This function generates the next clause, based on the given literals and a
 * canonical representation of the last clause. Increments prev_rep.
 * @param all_literals: list of literals of the language
 * @param prev_rep: list of 0/1 indicating which literals of all_literals are
 * in the last generated clause
 * 
 * @returns the next clause
 * 
 */
literalslist* next_clause_rep(literalslist* all_literals, literalslist **prev_rep) {
    literalslist *ret = NULL;
    literalslist *cur_rep = *prev_rep;
    literalslist *cur_lit = all_literals;

    // Add only the literals which have 1 in the representation
    while(cur_rep){
        if (cur_rep->literal == 1) {
            literalslist *aux = malloc(sizeof(literalslist));
            aux->literal = cur_lit->literal;
            aux->next = ret;
            ret = aux;
        }
        cur_rep = cur_rep->next;
        cur_lit = cur_lit->next;
    }

    cur_rep = *prev_rep;

    // Increments representation
    // If there is an overflow on the representation, deletes it
    while(cur_rep){
        if(cur_rep->literal == 0) {
            cur_rep->literal = 1;
            break;
        } else if(cur_rep->literal == 1) {
            if(cur_rep->next == NULL) {
                free_disjunction(*prev_rep);
                *prev_rep = NULL;
                break;
            }
            cur_rep->literal = 0;
        }
        cur_rep = cur_rep->next;
    }
    

    return ret;
}

// Unknown clauses

typedef struct cl_list {
    tclause *clause;
    struct cl_list *next;
    struct cl_list *prev;
} cl_list;


void rm_cl_from_unknown(cl_list **unk, cl_list **c_cl, cl_list **end) {
    cl_list *this = *c_cl;
    if(this == *unk || this == *end) {
        if(this == *unk) {
            *unk = this->next;
            if(*unk) (*unk)->prev = NULL;
        }
        if(this == *end) {
            *end = this->prev;
           if(*end) (*end)->next = NULL;
        }
        *c_cl = this->next;
        free(this->clause);
        free(this);
    } else {
        if(this->prev) this->prev->next = this->next;
        if(this->next) this->next->prev = this->prev;
        *c_cl = this->next;
        free(this->clause);
        free(this);
    }
}

cl_list *add_cl_unknown(tclause *clause, cl_list **unk, cl_list **c_cl,
                        cl_list **end) {
    cl_list *tmp_cl = malloc(sizeof(cl_list));
    tmp_cl->clause = clause;
    cl_list *aux = *end;
    if(*unk == NULL) *unk = tmp_cl;
    if(*end == NULL) *end = tmp_cl;
    tmp_cl->prev = aux;
    tmp_cl->next = NULL;
    if(aux) aux->next = tmp_cl;
    if(*c_cl == NULL) *c_cl = tmp_cl;
    *end = tmp_cl;

    return tmp_cl;
}

// Processing loop


int processing (int cycle) {

    literalslist *literals_1 = get_literals(m_sos_2);
    literalslist *literals_2 = get_literals(m_sos_1);

    cl_list *cur_cl_1 = NULL;
    cl_list *end_cl_1 = NULL;
    cl_list **unknown_1 = NULL;
    cl_list *cur_cl_2 = NULL;
    cl_list *end_cl_2 = NULL;
    cl_list **unknown_2 = NULL;

    unknown_1 = malloc(sizeof(cl_list *));
    *unknown_1 = NULL;
    unknown_2 = malloc(sizeof(cl_list *));
    *unknown_2 = NULL;

    literalslist *temp = NULL;
    literalslist *bool_list_1 = NULL;
    literalslist *bool_list_2 = NULL;
    literalslist **rep_1 = &bool_list_1;
    literalslist **rep_2 = &bool_list_2;

    hml_pclauses *known_1 = NULL;
    hml_pclauses *known_2 = NULL;

    int no_gen_1c = 1;
    int no_gen_2c = 1;

    int agent = 1;
    literalslist *aux = NULL;
    int negs = 0;

    // Generate the representation list of the literals in C_1
    literalslist *c = NULL;
    c = literals_1;
    while(c != NULL) {
        temp = malloc(sizeof(literalslist));
        temp->literal = 0;
        temp->next = bool_list_1;
        bool_list_1 = temp;
        c = c->next;
    }

    // Generate the representation list of the literals in C_2
    c = literals_2;

    while(c != NULL) {
        temp = malloc(sizeof(literalslist));
        temp->literal = 0;
        temp->next = bool_list_2;
        bool_list_2 = temp;
        c = c->next;
    }

    next_clause_rep(literals_1, rep_1);
    next_clause_rep(literals_2, rep_2);

    int all_generated_2 = 0;
    int all_generated_1 = 0;

    literalslist *cur = NULL; 
    tclause *tmp = NULL;
    cur = next_clause_rep(literals_2, rep_2);
    if(cur) {
        tmp = to_tclause(cur, QUERY2);
        add_cl_unknown(tmp, unknown_2, &cur_cl_2, &end_cl_2); //cur_cl_2 initialized
    }
    
    cur = next_clause_rep(literals_1, rep_1);
    if(cur) {
        tmp = to_tclause(cur, QUERY1);
        add_cl_unknown(tmp, unknown_1, &cur_cl_1, &end_cl_1); //cur_cl_1 initialized
    }

    cur = NULL;

    do {

        no_gen_1c = 1;
        while(cur_cl_2) {

            // Negates the literals, preserving the original clause for querying
            literalslist *inv_cur = NULL;
            if(cur_cl_2) cur = cur_cl_2->clause->right;
            c = cur; 
            while(c != NULL) {
                literalslist *c_inv = NULL;
                c_inv = malloc(sizeof(literalslist));
                c_inv->literal = -c->literal;
                c_inv->next = inv_cur;
                inv_cur = c_inv;
                c = c->next;
            }
            c = cur;

            // Counts the number of negative modal clauses
            agent = 1;
            aux = NULL;
            negs = 0;
            for(agent = 1; agent < numagents; agent++) {
                    aux = inv_cur;
                    while (aux != NULL)
                    {
                        clauseslist *negcandidates =
                        build_modal_candidates(m_sos_1, agent, 1,
                        MNEGATIVE, aux->literal);
                        clauseslist *clause_list_p = negcandidates;
                        while(clause_list_p) {
                            if(clause_list_p->clause_pointer->type ==
                            MNEGATIVE) negs++;

                            clause_list_p = clause_list_p->next;
                        }
                        aux = aux->next;
                    }

            }

            // If there is more than one, skips this guess
            if(negs > 1) {
                cur = next_clause_rep(literals_2, rep_2);
                if(cur) {
                    tclause *tmp = to_tclause(cur, QUERY1);
                    if(cur_cl_1->next == NULL) {
                        add_cl_unknown(tmp, unknown_2, &cur_cl_2, &end_cl_2);
                    } else{
                        add_cl_unknown(tmp, unknown_2, &cur_cl_2, &end_cl_2);
                        cur_cl_2 = cur_cl_2->next;

                    }
                    continue;
                } else {
                    all_generated_2 = 1;
                    continue;
                }
            }

            tclause *inverted = to_tclause(inv_cur, QUERY2);

            // Check if query-clause is already subsumed
            int subs = 0;
            if(known_2) subs = forward_subsumed(&inverted, known_2);
            if(subs) {
                if(print_subsumed) print_clause(inverted);
                fflush(stdout);
                free_disjunction(inv_cur);
                free(inverted);
                if(!all_generated_2) {
                    cur = next_clause_rep(literals_2, rep_2);
                    if(!cur) all_generated_2 = 1;
                }
                rm_cl_from_unknown(unknown_2, &cur_cl_2, &end_cl_2); // This advances cur_cl_2
                continue;
            }


            // Query the component, add resolvents
            int ret = make_query(2, cur);
            if(print_queries) print_query(inverted, 2, !ret);
            fflush(stdout);
            inverted->just->rule = LRES2;
            if(ret == 0) {
                insert_clause_set(&inverted, inverted->type, SOS);
                do_gen1(inverted->number, m_sos_1);
                do_gen3(inverted->number, m_sos_1);
                delete_clause_prop(&l_sos, inverted);
                insert_clause_set(&inverted, LITERAL, SOS_2);
                int generated = move_pclauses_from_set(&l_sos, SOS_1);
                if(generated) no_gen_1c = 0;
                append_clause_set(&inverted, &known_2);
                rm_cl_from_unknown(unknown_2, &cur_cl_2, &end_cl_2);
            } else {
                free_disjunction(inv_cur);
                free(inverted);
            }

            // If there are still clauses to be generated, do it. 
            if(!all_generated_2) {
                cur = next_clause_rep(literals_2, rep_2);
                if(cur) {
                    tclause *tmp = to_tclause(cur, QUERY2);
                    add_cl_unknown(tmp, unknown_2, &cur_cl_2, &end_cl_2);
                    // If the generated clause was on the end of the list, continue so it is the next current clause
                    if(cur_cl_2->next == NULL) continue;
                } else {
                    all_generated_2 = 1;
                }
            }
            if(cur_cl_2) cur_cl_2 = cur_cl_2->next;
        }

        no_gen_2c = 1;
        while(cur_cl_1) {

            // Negates the literals, preserving the original clause for querying
            literalslist *inv_cur = NULL;
            if(cur_cl_1) cur = cur_cl_1->clause->right;
            c = cur; 
            while(c != NULL) {
                literalslist *c_inv = NULL;
                c_inv = malloc(sizeof(literalslist));
                c_inv->literal = -c->literal;
                c_inv->next = inv_cur;
                inv_cur = c_inv;
                c = c->next;
            }
            c = cur;

            // Counts the number of negative modal clauses
            agent = 1;
            aux = NULL;
            negs = 0;
            for(agent = 1; agent < numagents; agent++) {
                    aux = inv_cur;
                    while (aux != NULL)
                    {
                        clauseslist *negcandidates =
                        build_modal_candidates(m_sos_2, agent, 1,
                        MNEGATIVE, aux->literal);
                        clauseslist *clause_list_p = negcandidates;
                        while(clause_list_p) {
                            if(clause_list_p->clause_pointer->type ==
                            MNEGATIVE) negs++;

                            clause_list_p = clause_list_p->next;
                        }
                        aux = aux->next;
                    }

            }

            // If there is more than one, skips this guess
            if(negs > 1) {
                cur = next_clause_rep(literals_1, rep_1);
                if(cur) {
                    tclause *tmp = to_tclause(cur, QUERY1);
                    if(cur_cl_1->next == NULL) {
                        add_cl_unknown(tmp, unknown_1, &cur_cl_1, &end_cl_1);
                    } else{
                        add_cl_unknown(tmp, unknown_1, &cur_cl_1, &end_cl_1);
                        cur_cl_1 = cur_cl_1->next;

                    }
                    continue;
                } else {
                    all_generated_1 = 1;
                    continue;
                }
            }

            tclause *inverted = to_tclause(inv_cur, QUERY1);

            // Check if query-clause is already subsumed
            int subs = 0;
            if(known_1) subs = forward_subsumed(&inverted, known_1);
            if(subs) {
                if(print_subsumed) print_clause(inverted);
                free_disjunction(inv_cur);
                free(inverted);
                if(!all_generated_1) {
                    cur = next_clause_rep(literals_1, rep_1);
                    if(!cur) all_generated_1 = 1;
                }
                rm_cl_from_unknown(unknown_1, &cur_cl_1, &end_cl_1); // This advances cur_cl_1
                continue;
            }


            // Query the component, add resolvents
            int ret = make_query(1, cur);
            if(print_queries) print_query(inverted, 1, !ret);
            fflush(stdout);
            inverted->just->rule = LRES1;
            if(ret == 0) {
                insert_clause_set(&inverted, inverted->type, SOS);
                do_gen1(inverted->number, m_sos_2);
                do_gen3(inverted->number, m_sos_2);
                delete_clause_prop(&l_sos, inverted);
                insert_clause_set(&inverted, LITERAL, SOS_1);
                int generated = move_pclauses_from_set(&l_sos, SOS_2);
                if(generated) no_gen_2c = 0;
                append_clause_set(&inverted, &known_1);
                rm_cl_from_unknown(unknown_1, &cur_cl_1, &end_cl_1);
            } else {
                free_disjunction(inv_cur);
                free(inverted);
            }

            // If there are still clauses to be generated, do it. 
            if(!all_generated_1) {
                cur = next_clause_rep(literals_1, rep_1);
                if(cur) {
                    tclause *tmp = to_tclause(cur, QUERY1);
                    add_cl_unknown(tmp, unknown_1, &cur_cl_1, &end_cl_1);
                    // If the generated clause was on the end of the list, continue so it is the next current clause
                    if(cur_cl_1->next == NULL) continue;
                } else {
                    all_generated_1 = 1;
                }
            } 
            if(cur_cl_1) cur_cl_1 = cur_cl_1->next;
            
        }

        int consistent = make_query(1, NULL);
        if(!consistent) {
            literalslist *f = malloc(sizeof(literalslist));
            f->literal = CFALSE;
            f->next = NULL;
            tclause *ft = to_tclause(f, LRES1);
            insert_clause_set(&ft, LITERAL, SOS_1);
            break;
        }

        cur_cl_1 = *unknown_1;
        cur_cl_2 = *unknown_2;

    } while((no_gen_1c == 0 || no_gen_2c == 0) && (*unknown_1 || *unknown_2));

    if(literals_1) free_disjunction(literals_1);
    if(literals_2) free_disjunction(literals_2);

    return 0;
}

