#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h> 
#include "uthash.h"
#include "tree.h"
#include "symbol_table.h"
#include "clauses.h"

#define MAX_QUERY 1000000
#define MAX_SIZE MAX_QUERY + 100
#define MAX_OUT 20

extern hml_mclauses *m_sos_1;
extern hml_mclauses *m_usable_1;
extern hml_mclauses *m_sos_2;
extern hml_mclauses *m_usable_2;
extern hml_pclauses *l_sos_1;
extern hml_pclauses *l_usable_1;
extern hml_pclauses *l_sos_2;
extern hml_pclauses *l_usable_2;

extern void s_print_disjunction(char** string, literalslist *s);
extern void s_print_conjunction(char** string, literalslist *s);
extern void s_print_clause(char** s, tclause *clause);
extern tclause *get_clause (int clause);
extern void bprint(char **buffer, char *string);

extern agent_node *find_agent (int id);

#define READ 0
#define WRITE 1

// 0 = Normal; 1 = T; 2 = four; 3 = T + four;
int component_box_type[2] = { 0, 0 };

pid_t popen2(const char **command, int *infp, int *outfp, int *child_infp, int *child_outfp) {

    int p_stdin[2], p_stdout[2];
    pid_t pid;

    if (pipe(p_stdin) != 0 || pipe(p_stdout) != 0)
        return -1;

    pid = fork();

    if (pid < 0)
        return pid;
    else if (pid == 0)
    {
        close(p_stdin[WRITE]);
        dup2(p_stdin[READ], READ);
        close(p_stdout[READ]);
        dup2(p_stdout[WRITE], WRITE);

        execvp(*command, (char * const *)command);
        perror("execvp");
    }

    if (infp == NULL || child_infp == NULL) {
        close(p_stdin[WRITE]);
        close(p_stdin[READ]);
    } else {
        *infp = p_stdin[WRITE];
        *child_infp = p_stdin[READ];
    }

    if (outfp == NULL || child_outfp == NULL) {
        close(p_stdout[READ]);
        close(p_stdout[WRITE]);
    } else {
        *outfp = p_stdout[READ];
        *child_outfp = p_stdout[WRITE];
    }

    return pid;
}

int query_prover(char* query, int component) {
    char result[MAX_OUT];

    if(component != 1 && component != 2) {
        printf("\nERROR: Invalid component");
        exit(1);
    }

    char *ksp[] = {"./comp", "-g", "-fsub", "-snf+","-neg",NULL};
    
    int *in = malloc(sizeof(int));
    int *out = malloc(sizeof(int));
    int *child_in = malloc(sizeof(int));
    int *child_out = malloc(sizeof(int));

    popen2((const char **) ksp, in, out, child_in, child_out);

    FILE *fdin = fdopen(*in, "a");
    FILE *fdout = fdopen(*out, "r");

    if(component_box_type[component - 1] == 0 ) {
        fprintf(fdin, "sos(formulas). %s. end_of_list.\n", query);
    } else if(component_box_type[component - 1] == 1)  {
        fprintf(fdin, "set(box, t). sos(formulas). %s. end_of_list.\n", query);
    } else if(component_box_type[component - 1] == 2)  {
        fprintf(fdin, "set(box, four). sos(formulas). %s. end_of_list.\n",
        query);
    } else if(component_box_type[component - 1] == 3)  {
        fprintf(fdin, "set(box, t, four). sos(formulas). %s. end_of_list.\n",
        query);
    } else {
        printf("\nERROR: invalid box type");
        exit(1);
    }

    fflush(fdin);
    close(*in);

    fgets(result, 4, fdout);

    int status;
    wait(&status);
    close(*out);

    close(*child_in);
    close(*child_out);

    free(in);
    free(out);
    free(child_in);
    free(child_out);

    if (strncmp("Sat", result, 3) == 0) {
        return 1;
    } else if (strncmp("Uns", result, 3) == 0) {
        return 0;
    } else {
        return -1;
    }


}


void s_print_mclauses (char **s, hml_mclauses *set) {
    hml_mclauses *aux = set;
    hagentclauses *hag = NULL;
    htypeclauses *htc = NULL;
    hmaxlitclauses *hmaxlit = NULL;
    int first = 1;

    for (aux=set; aux != NULL; aux=aux->hml.next)
        for (hag=aux->id; hag != NULL; hag=hag->hid.next)
            for (htc=hag->type; htc != NULL; htc=htc->htype.next) {
                agent_node *a = find_agent(hag->id);
                if(a != NULL && strstr(a->name, "dbox") == NULL) {
                    for (hmaxlit = htc->lit; hmaxlit != NULL; hmaxlit=hmaxlit->hmax.next) {
                        clauseslist *l = hmaxlit->list;
                        while (l != NULL) {
                            if(first) {
                                first = 0;
                            } else {
                                bprint(s, " & ");
                            }
                            tclause *clause = get_clause(l->clause_number);
                            bprint(s, "(");
                            s_print_clause(s, clause);
                            bprint(s, ")");
                            l = l->next;
                        }
                    }
                }
            }
}

void s_print_pclauses (char **s, hml_pclauses *set) {
    hml_pclauses *aux = set;
    hmaxclauses *hmc = NULL;
    hsizeclauses *hsc = NULL;  

    int first = 1;

    for (aux=set; aux != NULL; aux=aux->hml.next)
        for (hmc=aux->max; hmc != NULL; hmc=hmc->hmax.next)
            for (hsc=hmc->size; hsc != NULL; hsc=hsc->hsize.next) {
                clauseslist *l = hsc->list;
                while (l != NULL) {
                    if(first) {
                        first = 0;
                    } else {
                        bprint(s, " & ");
                    }

                    tclause *clause = get_clause(l->clause_number);
                    bprint(s, "(");
                    s_print_clause(s, clause);
                    bprint(s, ")");
                    l = l->next;
                }
            }
}

int make_query(int component, literalslist *literals) {
    char **buffer = malloc(sizeof(char *));
    (*buffer) = malloc(sizeof(char)*MAX_QUERY);
    char *query = *buffer; 

    hml_mclauses *m_sos = NULL;
    hml_mclauses *m_usable = NULL;
    hml_pclauses *l_sos = NULL;
    hml_pclauses *l_usable = NULL;

    if(component == 1) {
        m_sos = m_sos_1;
        m_usable = m_usable_1;
        l_sos = l_sos_1;
        l_usable = l_usable_1;
    } else {
        m_sos = m_sos_2;
        m_usable = m_usable_2;
        l_sos = l_sos_2;
        l_usable = l_usable_2;
    }



    char *old = *buffer;
    s_print_mclauses(buffer, m_sos);
    if(old != *buffer && (m_usable || l_sos || l_usable || literals)) bprint(buffer, " & ");
    old = *buffer;

    s_print_mclauses(buffer, m_usable);
    if(old != *buffer && (l_sos || l_usable || literals)) bprint(buffer, " & ");
    old = *buffer;

    s_print_pclauses(buffer, l_sos);
    if(old != *buffer && (l_usable || literals)) bprint(buffer, " & ");
    old = *buffer;

    s_print_pclauses(buffer, l_usable);
    if(literals) {
        if(old != *buffer) bprint(buffer, " & ");
        old = *buffer;

        bprint(buffer, "(");
        s_print_conjunction(buffer, literals);
        bprint(buffer, ")");
    }

    int ret = query_prover(query, component);
    free(query);
    free(buffer);

    return ret;
}
